# The MIT License (MIT)
# Copyright (c) 2025, 2026 The Galactipy Contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

"""Verify if a candidate version tag refers to a stable or unstable release."""

from typing import Annotated

import typer

from quantum_foam import check_stable_version

check_app = typer.Typer(rich_markup_mode="rich")

HELP_MSG = (
    ":white_check_mark: "
    "Verify if a candidate version tag refers to a stable or unstable release."
)


@check_app.command(help=HELP_MSG)
def check(
    tag: Annotated[
        str,
        typer.Argument(help=":bookmark: A candidate tag version to extract info from."),
    ],
) -> None:
    """Verify if a candidate version tag refers to a stable or unstable release."""
    exit_code = 0 if check_stable_version(tag) else 1

    raise typer.Exit(exit_code)



# The MIT License (MIT)
# Copyright (c) 2025, 2026 The Galactipy Contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.


# UPDATEME With subcommand apps in `cli/commands/`, see documentation at:
# https://typer.tiangolo.com/tutorial/
# See recommended configuration for multicommand applications at:
# https://typer.tiangolo.com/tutorial/one-file-per-command/#main-module-mainpy
"""Convert version tags between different schemas."""

from typing import Annotated

import typer
from rich.console import Console

from quantum_foam import __version__
from quantum_foam.cli.commands.check_command import check_app
from quantum_foam.cli.commands.pvi_command import pvi_app
from quantum_foam.cli.commands.pypi_command import pypi_app
from quantum_foam.cli.commands.semver_command import semver_app
from quantum_foam.cli.styling import AppCustomThemes

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
app.add_typer(pypi_app)
app.add_typer(semver_app)
app.add_typer(pvi_app)
app.add_typer(check_app)


def version_callback(print_version: bool) -> None:
    """Print the program version in a Rich console with the Noctis theme."""
    if print_version:
        Console(theme=AppCustomThemes.NOCTIS).print(
            f":package:[declaration]Quantum Foam[/] [bold fstring]{__version__}[/]"
        )

        raise typer.Exit


@app.callback()
def main(
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            "-v",
            help=":bulb: Print the current version of this program and exit.",
            callback=version_callback,
            is_eager=True,
        ),
    ] = False,
) -> None:
    """Convert version tags between different schemas.

    See below for commands and options.
    """
    pass



# The MIT License (MIT)
# Copyright (c) 2025, 2026 The Galactipy Contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.


"""Return the version identifier for a candidate version tag."""

from typing import Annotated

import typer
from rich import print as rich_print

from quantum_foam import extract_pvi

pvi_app = typer.Typer(rich_markup_mode="rich")


@pvi_app.command()
def pvi(
    tag: Annotated[
        str,
        typer.Argument(help=":bookmark: A candidate tag version to extract info from."),
    ],
    compound: Annotated[
        bool,
        typer.Option(
            "--compound",
            "-c",
            help=':link: Include "release" as the head noun of the label',
        ),
    ] = False,
    capitalize: Annotated[
        bool,
        typer.Option(
            "--capitalize",
            "-C",
            help=":gem: Capitalise the first letter of the version identifier.",
        ),
    ] = False,
) -> None:
    """:dart: Extract the tag version identifier."""
    pvi = extract_pvi(tag).capitalize() if capitalize else extract_pvi(tag)

    if compound and "release" not in pvi:
        rich_print(f"{pvi} release")

    else:
        rich_print(pvi)



# The MIT License (MIT)
# Copyright (c) 2025, 2026 The Galactipy Contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.


"""Perform conversions between versioning schemas."""

import re

from packaging.version import VERSION_PATTERN, InvalidVersion
from packaging.version import Version as PyPI
from semver.version import Version as Semver


def _check_prefixed_version(candidate_tag: str) -> str:
    """Check if a candidate version tag is prefixed by "v".

    As SemVer specification is strict and disavows usage of prefixed versions,
    candidate tags must be cleaned before being parsed.
    """  # noqa: DOC201
    if candidate_tag.casefold().startswith("v"):
        return candidate_tag.lstrip("vV")

    return candidate_tag


def convert_to_semver(candidate_tag: str) -> str:
    """Convert a candidate tag to its equivalent semantic version.

    Parameters
    ----------
    candidate_tag : str
        A string with a possible version tag to be converted.

    Returns
    -------
    str
        The string for the candidate tag in semantic version notation.

    Raises
    ------
    InvalidVersion
        If the provided string can not be converted to SemVer.
    """
    clean_tag = _check_prefixed_version(candidate_tag)

    try:
        return str(Semver.parse(clean_tag))

    except ValueError:
        msg = f'"{clean_tag}" is not a valid semantic version'
        raise InvalidVersion(msg) from None


def convert_to_pypi(candidate_tag: str) -> str:
    """Convert a candidate tag to the PyPI version specifier.

    Parameters
    ----------
    candidate_tag : str
        A string with a possible version tag to be converted.

    Returns
    -------
    str
        The string for the candidate tag in the PyPI-compliant schema.

    Raises
    ------
    InvalidVersion
        If the provided string can not be converted to the PyPI version specifier.
    """
    try:
        return str(PyPI(candidate_tag))

    except InvalidVersion:
        msg = f'"{candidate_tag}" does not adhere to the PyPI version specifier'
        raise InvalidVersion(msg) from None


def _get_version_stage(pre: str | None, post: str | None, dev: str | None) -> str:
    """Determine the version stage from version components.

    Parameters
    ----------
    pre : str, None
        Pre-release identifier obtained from the PyPI version (e.g., "a", "b", "rc").
    post : str, None
        Post-release identifier.
    dev : str, None
        Development release identifier.

    Returns
    -------
    str
        The version stage name.
    """
    # Map pre-release identifiers to stage names
    pre_release_map = {"a": "alpha", "b": "beta", "rc": "pre-release"}

    if pre in pre_release_map:
        return pre_release_map[pre]

    if post is not None:
        return "post-release"

    if dev is not None:
        return "development"

    return "release"


def extract_pvi(candidate_tag: str) -> str:
    """Extract the Package Version Identifier (PVI) from a candidate tag.

    Parameters
    ----------
    candidate_tag : str
        A string with a possible version tag.

    Returns
    -------
    str
        The version stage identifier.
    """
    r = re.compile(VERSION_PATTERN, re.VERBOSE)
    pypi_version = convert_to_pypi(candidate_tag)

    match = r.match(pypi_version).groupdict()  # type: ignore[union-attr]

    return _get_version_stage(
        match.get("pre_l"), match.get("post_l"), match.get("dev_l")
    )


def check_stable_version(candidate_tag: str) -> bool:
    """Verify if the major version is greater than zero (i.e., a stable version).

    Parameters
    ----------
    candidate_tag : str
        A string with a possible version tag.

    Returns
    -------
    bool
        Whether the version parsed is a stable release or not.

    Raises
    ------
    InvalidVersion
        If the provided string can not be converted to SemVer.
    """
    clean_tag = _check_prefixed_version(candidate_tag)

    try:
        semver = Semver.parse(clean_tag)

    except ValueError:
        msg = f'"{clean_tag}" is not a valid semantic version'
        raise InvalidVersion(msg) from None

    else:
        return semver.major > 0



# The MIT License (MIT)
# Copyright (c) 2025, 2026 The Galactipy Contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.


"""Specify Rich styles to render text output in the terminal.

Themes defined in this module can be added to the `AppCustomStyles` class and used in
conjunction with `rich.console.Console` to override the default terminal theme.

See https://rich.readthedocs.io/en/stable/style.html#style-themes for more information.

"""

from dataclasses import dataclass

from rich.theme import Theme

# Noctis VS Code theme text colours, full colorscheme definition at:
# https://github.com/liviuschera/noctis
NOCTIS_THEME = Theme(
    {
        "string": "#49e9a6",
        "fstring": "#16b673",
        "comment": "#5b858b",
        "function": "#16a6b6",
        "method": "#49d6e9",
        "standout": "#49ace9",
        "number": "#7060eb",
        "keyword": "#df769b",
        "declaration": "#e66533",
        "property": "#d67e5c",
        "constant": "#d5971a",
        "variable": "#e4b781",
        "bold string": "bold #49e9a6",
        "bold fstring": "bold #16b673",
        "bold comment": "bold #5b858b",
        "bold function": "bold #16a6b6",
        "bold method": "bold #49d6e9",
        "bold standout": "bold #49ace9",
        "bold number": "bold #7060eb",
        "bold keyword": "bold #df769b",
        "bold declaration": "bold #e66533",
        "bold property": "bold #d67e5c",
        "bold constant": "bold #d5971a",
        "bold variable": "bold #e4b781",
    }
)


@dataclass(frozen=True)
class AppCustomThemes:
    """Define custom Rich themes for console output with Quantum Foam."""

    NOCTIS: Theme = NOCTIS_THEME
